/* This file is part of Idest.
   Copyright (C) 2009-2011 Sergey Poznyakoff
 
   Idest is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 3, or (at your option)
   any later version.
 
   Idest is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
 
   You should have received a copy of the GNU General Public License
   along with Idest.  If not, see <http://www.gnu.org/licenses/>. */

#include "idest.h"
#include <signal.h>

void
set_frame_value(struct id3_frame *frame, const char *value)
{
	union id3_field *field;
	enum id3_field_textencoding encoding =
		(latin1_option ? ID3_FIELD_TEXTENCODING_ISO_8859_1 
		               : ID3_FIELD_TEXTENCODING_UTF_8);
	id3_ucs4_t *ucs4;
	
	switch (frame_id(frame->id)) {
	case item_title:
	case item_artist:
	case item_album:
	case item_year:
	case item_track:
	case item_genre:
		field = id3_frame_field(frame, 0);
		id3_field_settextencoding(field, encoding);

		if (latin1_option)
			ucs4 = id3_latin1_ucs4duplicate(
				     (const id3_latin1_t *) value);
		else
			ucs4 = id3_utf8_ucs4duplicate(
				     (const id3_utf8_t *)value);
		
		field = id3_frame_field(frame, 1);
		id3_field_setstrings(field, 1, &ucs4);
		free(ucs4);
		break;
		
	case item_comment:
		/* Field 0: Encoding */
		field = id3_frame_field(frame, 0);
		id3_field_settextencoding(field, encoding);
		/* FIXME: Field 1: Language */
		/* FIXME: Field 2: String */
		/* Field 3: Comment */
		field = id3_frame_field(frame, 3);
		if (latin1_option)
			ucs4 = id3_latin1_ucs4duplicate(
				     (const id3_latin1_t *) value);
		else
			ucs4 = id3_utf8_ucs4duplicate(
				     (const id3_utf8_t *)value);
		id3_field_setfullstring(field, ucs4);
		free(ucs4);
		break;
		
	default:
		error(1, 0, "don't know how to set field %s", frame->id);
	}
}

void
safe_id3_file_update_and_close(struct id3_file *file)
{
	sigset_t set, oldset;

	sigemptyset(&set);
	sigaddset(&set, SIGINT);
	sigaddset(&set, SIGTERM);

	sigprocmask(SIG_BLOCK, &set, &oldset);
	id3_file_update(file);
	id3_file_close(file);
	sigprocmask(SIG_SETMASK, &oldset, NULL);
}

void
set_tags(const char *name)
{
	struct id3_file *file;
	struct id3_tag *tag;
	int modified = 0;
	
	file = id3_file_open(name, ID3_FILE_MODE_READWRITE);
	if (!file)
		error(1, errno, "cannot open file %s", name);
	tag = id3_file_tag(file);
	if (!tag) 
		abort(); /* FIXME */

	if (ed_list) {
		gl_list_iterator_t itr;
		const void *p;
		
		itr = gl_list_iterator(ed_list);
		while (gl_list_iterator_next(&itr, &p, NULL)) {
			const struct ed_item *item = p;
			struct id3_frame *frame
				= id3_tag_findframe(tag, item->id, 0);
			if (!frame) {
				frame = id3_frame_new(item->id);
				if (id3_tag_attachframe(tag, frame))
					error(1, 0, "cannot attach new frame");
			}
			set_frame_value(frame, item->v.value);
			modified |= 1;
		}
		gl_list_iterator_free(&itr);
	}

	/* FIXME */
	modified |= guile_transform(name, tag);

	switch (convert_version) {
	case 1:
		id3_tag_options(tag,
				ID3_TAG_OPTION_NO_ID3V2,
				ID3_TAG_OPTION_NO_ID3V2);
		version_option = 1;
		modified |= 1;
		break;

	case 2:
		id3_tag_options(tag, ID3_TAG_OPTION_ID3V1, 0);
		version_option = 0;
		modified |= 1;
		break;
	}
	
	if (version_option == 1) {
		id3_tag_options(tag, ID3_TAG_OPTION_ID3V1,
				ID3_TAG_OPTION_ID3V1);
		modified |= 1;
	}
	if (modified)
		safe_id3_file_update_and_close(file);
	else
		id3_file_close(file);
}

void
del_tags(const char *name)
{
	struct id3_file *file;
	struct id3_tag *tag;
	
	file = id3_file_open(name, ID3_FILE_MODE_READWRITE);
	if (!file)
		error(1, errno, "cannot open file %s", name);
	tag = id3_file_tag(file);
	if (!tag) 
		abort(); /* FIXME */
	id3_tag_options(tag,
			ID3_TAG_OPTION_ID3V1|ID3_TAG_OPTION_NO_ID3V2,
			ID3_TAG_OPTION_NO_ID3V2);
	safe_id3_file_update_and_close(file);
}


char *
idest_ucs4_cvt(id3_ucs4_t const *ucs4)
{
	if (latin1_option) 
		return (char*)id3_ucs4_latin1duplicate(ucs4);
	else
		return (char*)id3_ucs4_utf8duplicate(ucs4);
}

static void
add_stringlist(gl_list_t list, struct id3_frame *frame,
	       union id3_field *field)
{
	unsigned i, nstrings = id3_field_getnstrings(field);
	for (i = 0; i < nstrings; i++) {
		id3_ucs4_t const *ucs4;
		char *str;
		
		ucs4 = id3_field_getstrings(field, i);
		if (!ucs4)
			continue;
		if (strcmp(frame->id, ID3_FRAME_GENRE) == 0)
			ucs4 = id3_genre_name(ucs4);
		str = idest_ucs4_cvt(ucs4);
		gl_list_add_last(list, str);
	}
}

static void
add_field(gl_list_t list, struct id3_frame *frame, union id3_field *field)
{
	id3_ucs4_t const *ucs4;
	char *str;
	
	switch (id3_field_type(field)) {
	case ID3_FIELD_TYPE_TEXTENCODING:
		break;
	case ID3_FIELD_TYPE_LATIN1:
	case ID3_FIELD_TYPE_LATIN1FULL:
	case ID3_FIELD_TYPE_LATIN1LIST:
		break;
	case ID3_FIELD_TYPE_STRING:
		break;
	case ID3_FIELD_TYPE_STRINGFULL:
		ucs4 = id3_field_getfullstring(field);
		str = idest_ucs4_cvt(ucs4);
		gl_list_add_last(list, str);
		break;
	case ID3_FIELD_TYPE_STRINGLIST:
		add_stringlist(list, frame, field);
		break;
	case ID3_FIELD_TYPE_LANGUAGE:
	case ID3_FIELD_TYPE_FRAMEID:
	case ID3_FIELD_TYPE_DATE:
	case ID3_FIELD_TYPE_INT8:
	case ID3_FIELD_TYPE_INT16:
	case ID3_FIELD_TYPE_INT24:
	case ID3_FIELD_TYPE_INT32:
	case ID3_FIELD_TYPE_INT32PLUS:
	case ID3_FIELD_TYPE_BINARYDATA:;
	}
}

static gl_list_t 
frame_to_list(struct id3_frame *frame)
{
	gl_list_t list;
	unsigned i;
	union id3_field *field;

	list = new_string_list(true);
	for (i = 0; (field = id3_frame_field(frame, i)); i++) 
		add_field(list, frame, field);
	return list;
}

static void
show_tags(struct id3_tag *tag)
{
	struct id3_frame *frame;
	unsigned i;

	for (i = 0; (frame = id3_tag_findframe(tag, NULL, i)); i++) {
		gl_list_t list = frame_to_list(frame);
		if (gl_list_size(list) > 0)
			ed_list_add_item(frame->id, list);
		else
			gl_list_free(list);
	}
	ed_list_print();
	ed_list_clear();
}

void
query_tags(const char *name)
{
	struct id3_file *file;
	struct id3_tag *tag;
	
	file = id3_file_open(name, ID3_FILE_MODE_READONLY);
	if (!file)
		error(1, errno, "cannot open file %s", name);

	tag = id3_file_tag(file);
	if (tag) {
		if (guile_list(name, tag) == 0)
			show_tags(tag);
	}
	
	id3_file_close(file);
}

void
del_id3v2(const char *name)
{
}
